
#include <cmath>
#include <cstdio>
#include <iostream>
#include <GL/gl.h>
#include <GL/glui.h>

using std::cos;
using std::sin;
using std::sqrt;
using std::cout;
using std::endl;


#define	TESSNUM		32
#define PIVAL		0.196349541
#define	CYLHEIGHT	2.0

GLfloat gamb[] = { 0.1, 0.1, 0.1, 1.0 };

GLfloat pos0[] = { -13.0, 5.0, 0.0, 1.0 };
GLfloat amb0[] = { 0.0, 0.05, 0.05, 1.0 };
GLfloat dif0[] = { 1.0, 0.0, 0.0, 1.0 };
GLfloat spec0[] = { 1.0, 1.0, 1.0, 1.0 };

GLfloat pos1[] = { 5.0, 5.0, -5.0, 1.0 };
GLfloat amb1[] = { 0.05, 0.0, 0.05, 1.0 };
GLfloat dif1[] = { 0.0, 1.0, 0.0, 1.0 };
GLfloat spec1[] = { 1.0, 1.0, 1.0, 1.0 };

GLfloat pos2[] = { -5.0, 5.0, -5.0, 1.0 };
GLfloat amb2[] = { 0.05, 0.05, 0.0, 1.0 };
GLfloat dif2[] = { 0.0, 0.0, 1.0, 1.0 };
GLfloat spec2[] = { 1.0, 1.0, 1.0, 1.0 };

GLfloat mamb[] = { 0.05, 0.05, 0.05, 1.0 };
GLfloat mdif[] = { 0.25, 0.5, 0.75, 1.0 };
GLfloat mspec[] = { 0.95, 0.85, 0.75, 1.0 };

GLfloat mreset[] = { 0.0f, 0.0f, 0.0f, 1.0f};

int bLighting = 0;
int bLight0 = 0;
int bLight1 = 0; 
int bLight2 = 0;

int bDiffuse = 1;
int bSpecular = 0;
int bAmbient = 1;
int bGlobalAmbient = 0;
int bSmooth = 1;

float shininess = 30.0f;
float rotMat[16];

float *tessBufferX;
float *tessBufferY;
float *tessBufferC;

// New Texturing Vars:
///////////////////////

int texturing = 0;
int separateSpec = 0;

int utiles = 1;
int vtiles = 1;
int uparam = 0;
int vparam = 0;
int texEnv = 0;

GLuint texName = 0;

// UI helpers

int colorModeButtonNum = 3;
char *colorModeButtonNames[] = { "Replace", "Modulate", "Decal" };
GLenum colorModeButtonVals[] = { GL_REPLACE, GL_MODULATE, GL_DECAL };

int texWrapButtonNum = 3;
char *texWrapButtonNames[] = { "Clamp", "Repeat", "Mirrored Repeat" };
GLenum texWrapButtonVals[] = { GL_CLAMP, GL_REPEAT, GL_MIRRORED_REPEAT };

////////////////////////


void buildUI(int mainWin)
{
	GLUI* uiWindow = GLUI_Master.create_glui("Lighting and texturing controls");

	GLUI_Panel *lightingPanel = uiWindow->add_panel("Lighting");

	uiWindow->add_checkbox_to_panel(lightingPanel, "Enabled", &bLighting);
	uiWindow->add_checkbox_to_panel(lightingPanel, "Light 0", &bLight0);
	uiWindow->add_checkbox_to_panel(lightingPanel, "Light 1", &bLight1);
	uiWindow->add_checkbox_to_panel(lightingPanel, "Light 2", &bLight2);

	GLUI_Panel *matPanel = uiWindow->add_panel("Material");

	uiWindow->add_checkbox_to_panel(matPanel, "Ambient term", &bAmbient);
	uiWindow->add_checkbox_to_panel(matPanel, "Diffuse term", &bDiffuse);
	uiWindow->add_checkbox_to_panel(matPanel, "Specular term", &bSpecular);

	GLUI_Spinner * spin = uiWindow->add_spinner_to_panel( matPanel
														, "Shininess"
														, GLUI_SPINNER_FLOAT
														, &shininess
													);

	spin->set_float_limits(0.5f, 100.0f);
	spin->set_speed(0.5f);

	uiWindow->add_checkbox("Global ambient term", &bGlobalAmbient);
	uiWindow->add_checkbox("Smooth shading", &bSmooth);

	uiWindow->add_separator();

	// Texturing GUI
	GLUI_Panel *texturePanel = uiWindow->add_panel("Texturing");
	
	uiWindow->add_checkbox_to_panel(texturePanel, "Enabled", &texturing);

	GLUI_Panel *colorModePanel = uiWindow->add_panel_to_panel(
									  texturePanel
									, "Color Mode (TexEnv)"
			);

	GLUI_RadioGroup *colorModeGroup = uiWindow->add_radiogroup_to_panel(
																colorModePanel
															  , &texEnv
			);

	for(int i = 0; i < colorModeButtonNum; ++i)
	{
		uiWindow->add_radiobutton_to_group( colorModeGroup
										  , colorModeButtonNames[i]
				);
	}

	GLUI_Panel *tilingU = uiWindow->add_panel_to_panel( texturePanel
													  , "Tiling U (TexParam)"
			);

	spin = uiWindow->add_spinner_to_panel( tilingU
										 , "Tiles"
										 , GLUI_SPINNER_INT
										 , &utiles
			);
	spin->set_int_limits(1, 25);
	spin->set_speed(0.005);
	
	GLUI_RadioGroup *uTileRadioGroup = uiWindow->add_radiogroup_to_panel(
											tilingU
										  , &uparam
			);

	for(int i = 0; i < texWrapButtonNum; ++i)
	{
		uiWindow->add_radiobutton_to_group( uTileRadioGroup
										  , texWrapButtonNames[i]
				);
	}

	GLUI_Panel *tilingV = uiWindow->add_panel_to_panel( texturePanel
													  , "Tiling V (TexParam)"
			);

	spin = uiWindow->add_spinner_to_panel( tilingV
										 , "Tiles"
										 , GLUI_SPINNER_INT
										 , &vtiles
			);
	spin->set_int_limits(1, 25);
	spin->set_speed(0.005);
	
	GLUI_RadioGroup *vTileRadioGroup = uiWindow->add_radiogroup_to_panel(
											tilingV
										  , &vparam
			);

	for(int i = 0; i < texWrapButtonNum; ++i)
	{
		uiWindow->add_radiobutton_to_group( vTileRadioGroup
										  , texWrapButtonNames[i]
				);
	}

	uiWindow->add_checkbox( "Separate Specular Color"
						  , &separateSpec
			);

	GLUI_Rotation *rot = uiWindow->add_rotation("Orientation", rotMat);
	rot->reset();

	uiWindow->set_main_gfx_window(mainWin);
}

void createTexture()
{
	std::FILE* f = std::fopen("texture.rawbga", "rb");

	unsigned width, height, length;

	std::fread(&width, sizeof(width), 1, f);
	std::fread(&height, sizeof(height), 1, f);

	length = 4*width*height;

	unsigned char *buffer = new unsigned char[length];

	std::fread(buffer, sizeof(unsigned char), length, f);

	std::fclose(f);

	glGenTextures(1, &texName);
	glBindTexture(GL_TEXTURE_2D, texName);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, texWrapButtonVals[uparam]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, texWrapButtonVals[vparam]);

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, colorModeButtonVals[texEnv]);

	glTexImage2D( GL_TEXTURE_2D
				, 0
				, GL_RGBA
				, width
				, height
				, 0
				, GL_RGBA
				, GL_UNSIGNED_BYTE
				, buffer
			);

	delete [] buffer;
}

void createTesselation()
{
	tessBufferX = new float[TESSNUM];
	tessBufferY = new float[TESSNUM];
	tessBufferC = new float[TESSNUM];

	float inv = 1.0/TESSNUM;

	for(int i = 0; i < TESSNUM; ++i)
	{
		tessBufferX[i] = cos(i*PIVAL);
		tessBufferY[i] = sin(i*PIVAL);
		tessBufferC[i] = i*inv;
	}
}

void drawCylinder()
{
	if(tessBufferX != 0)
	{
		glBegin(GL_TRIANGLE_STRIP);
		
		int i;
		float rev;

		for(i = 0; i < TESSNUM; ++i)
		{
			rev = 1.0 - tessBufferC[i];

			glColor3f(rev, rev, 1.0);
			glTexCoord2f(tessBufferC[i]*utiles, vtiles);
			glNormal3f(tessBufferX[i], tessBufferY[i], 0);
			glVertex3f(tessBufferX[i], tessBufferY[i], -CYLHEIGHT);

			glColor3f(rev, tessBufferC[i], 0);
			glTexCoord2f(tessBufferC[i]*utiles, 0);
			glNormal3f(tessBufferX[i], tessBufferY[i], 0);
			glVertex3f(tessBufferX[i], tessBufferY[i], CYLHEIGHT);
		}

		// close the cylinder.
		glColor3f(0.0, 0.0, 1.0);
		glTexCoord2f(utiles, vtiles);
		glNormal3d(tessBufferX[0], tessBufferY[0], 0);
		glVertex3d(tessBufferX[0], tessBufferY[0], -CYLHEIGHT);

		glColor3f(0.0, 1.0, 0.0);
		glTexCoord2f(utiles, vtiles);
		glNormal3d(tessBufferX[0], tessBufferY[0], 0);
		glVertex3d(tessBufferX[0], tessBufferY[0], CYLHEIGHT);

		glEnd();
	}
	else
	{
		createTesselation();
	}
}

void glInit()
{
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glEnable(GL_DEPTH_TEST);
	glShadeModel(GL_SMOOTH);
	glClearDepth(100.0);
	glFrontFace(GL_CW);

	// setup lights
	glLightfv(GL_LIGHT0, GL_POSITION, pos0);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, dif0);
	glLightfv(GL_LIGHT0, GL_SPECULAR, spec0);

	glLightfv(GL_LIGHT1, GL_POSITION, pos1);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, dif1);
	glLightfv(GL_LIGHT1, GL_SPECULAR, spec1);

	glLightfv(GL_LIGHT2, GL_POSITION, pos2);
	glLightfv(GL_LIGHT2, GL_DIFFUSE, dif2);
	glLightfv(GL_LIGHT2, GL_SPECULAR, spec2);


	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0, 1.0, 10.0, 100.0);

	glMatrixMode(GL_MODELVIEW);

	createTexture();
}

void reshape(int x, int y)
{
	glViewport(0, 0, x, y);
}

void display()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();
	gluLookAt(-13, 0, 0, 0, 0, 0, 0, 1, 0);
	glMultMatrixf(rotMat);


	// shading
	if(bSmooth)	glShadeModel(GL_SMOOTH);
	else		glShadeModel(GL_FLAT);

	// lighting
	if(bLighting)	glEnable(GL_LIGHTING);
	else			glDisable(GL_LIGHTING);

	// individual lights
	if(bLight0)		glEnable(GL_LIGHT0);
	else			glDisable(GL_LIGHT0);

	if(bLight1)		glEnable(GL_LIGHT1);
	else			glDisable(GL_LIGHT1);

	if(bLight2)		glEnable(GL_LIGHT2);
	else			glDisable(GL_LIGHT2);

	// material terms
	if(bDiffuse)	glMaterialfv(GL_FRONT, GL_DIFFUSE, mdif);
	else			glMaterialfv(GL_FRONT, GL_DIFFUSE, mreset);

	if(bSpecular)	glMaterialfv(GL_FRONT, GL_SPECULAR, mspec);
	else			glMaterialfv(GL_FRONT, GL_SPECULAR, mreset);

	if(bAmbient)	glMaterialfv(GL_FRONT, GL_AMBIENT, mamb);
	else			glMaterialfv(GL_FRONT, GL_AMBIENT, mreset);

	glMaterialf(GL_FRONT, GL_SHININESS, shininess);

	if(bGlobalAmbient)
	{
		glLightModelfv(GL_LIGHT_MODEL_AMBIENT, gamb);
	}
	else
	{
		glLightModelfv(GL_LIGHT_MODEL_AMBIENT, mreset);
	}

	// NEW TEXTURING CALLS
	///////////////////////
	if(separateSpec)
	{
		glLightModeli( GL_LIGHT_MODEL_COLOR_CONTROL
					 , GL_SEPARATE_SPECULAR_COLOR
				);
	}
	else
	{
		glLightModeli( GL_LIGHT_MODEL_COLOR_CONTROL
					 , GL_SINGLE_COLOR
				);
	}

	if(texturing)	glEnable(GL_TEXTURE_2D);
	else			glDisable(GL_TEXTURE_2D);

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, colorModeButtonVals[texEnv]);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, texWrapButtonVals[uparam]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, texWrapButtonVals[vparam]);

	//////////////////////
	drawCylinder();

	glutSwapBuffers();
}

int main(int argc, char *argv[])
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);

	glutInitWindowPosition(200, 200);
	glutInitWindowSize(500, 500);

	int win = glutCreateWindow("Lighting and Texturing Demo");

	glutDisplayFunc(display);
	glutReshapeFunc(reshape);

	GLUI_Master.set_glutIdleFunc(NULL);

	buildUI(win);

	glInit();

	glutMainLoop();

	return 0;
}
